#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import copy
import logging
import typing
import ccxt
import ccxt.pro as ccxt_pro
import ccxt.async_support as async_ccxt

import octobot_commons.time_frame_manager as time_frame_manager
import octobot_trading.constants as constants
import octobot_trading.enums as enums
import octobot_trading.errors as errors
import octobot_trading.exchanges.connectors.ccxt.enums as ccxt_enums
import octobot_trading.exchanges.connectors.ccxt.ccxt_clients_cache as ccxt_clients_cache
import octobot_trading.exchanges.config.proxy_config as proxy_config_import
import octobot_trading.exchanges.util.exchange_util as exchange_util


def create_client(
    exchange_class, exchange_manager, logger, options, headers,
    additional_config, should_authenticate, unauthenticated_exchange_fallback=None,
    keys_adapter=None
):
    """
    Exchange instance creation
    :return: the created ccxt (pro, async or sync) client
    """
    is_authenticated = False
    if not exchange_manager.exchange_only:
        # avoid logging version on temporary exchange_only exchanges
        exchange_type = exchange_util.get_exchange_type(exchange_manager)
        logger.info(f"Creating {exchange_class.__name__} {exchange_type.name} "
                    f"exchange with ccxt in version {ccxt.__version__}")
    if exchange_manager.ignore_config or exchange_manager.check_config(exchange_manager.exchange_name):
        try:
            auth_token_header_prefix = None
            key, secret, password, uid, auth_token = exchange_manager.get_exchange_credentials(
                exchange_manager.exchange_name
            )
            if keys_adapter:
                key, secret, password, uid, auth_token, auth_token_header_prefix = keys_adapter(
                    key, secret, password, uid, auth_token
                )
            if not (key and secret) and not exchange_manager.is_simulated and not exchange_manager.ignore_config:
                logger.warning(f"No exchange API key set for {exchange_manager.exchange_name}. "
                               f"Enter your account details to enable real trading on this exchange.")
            if should_authenticate and not exchange_manager.is_backtesting:
                client = exchange_class(_get_client_config(
                    options, headers, additional_config,
                    api_key=key, secret=secret, password=password, uid=uid,
                    auth_token=auth_token, auth_token_header_prefix=auth_token_header_prefix
                ))
                is_authenticated = True
                if exchange_manager.check_credentials:
                    client.checkRequiredCredentials()
            else:
                client = exchange_class(_get_client_config(options, headers, additional_config))
        except (ccxt.AuthenticationError, Exception) as e:
            if unauthenticated_exchange_fallback is None:
                return get_unauthenticated_exchange(
                    exchange_class, options, headers, additional_config, exchange_manager.proxy_config
                ), False
            return unauthenticated_exchange_fallback(e), False
    else:
        client = get_unauthenticated_exchange(
            exchange_class, options, headers, additional_config, exchange_manager.proxy_config
        )
        logger.error("configuration issue: missing login information !")
    client.logger.setLevel(logging.INFO)
    _use_proxy_if_necessary(client, exchange_manager.proxy_config)
    return client, is_authenticated


async def close_client(client):
    await client.close()
    client.markets = {}
    client.markets_by_id = {}
    client.ids = []
    client.last_json_response = {}
    client.last_http_response = ""
    client.last_response_headers = {}
    client.markets_loading = None
    client.currencies = {}
    client.baseCurrencies = {}
    client.quoteCurrencies = {}
    client.currencies_by_id = {}
    client.codes = []
    client.symbols = {}
    client.accounts = []
    client.accounts_by_id = {}
    client.ohlcvs = {}
    client.trades = {}
    client.orderbooks = {}


def get_unauthenticated_exchange(
    exchange_class, options, headers, additional_config, proxy_config: proxy_config_import.ProxyConfig
):
    client = exchange_class(_get_client_config(options, headers, additional_config))
    _use_proxy_if_necessary(client, proxy_config)
    return client


def set_sandbox_mode(exchange_connector, is_sandboxed):
    try:
        exchange_connector.client.setSandboxMode(is_sandboxed)
    except ccxt.NotSupported as e:
        default_type = exchange_connector.client.options.get('defaultType', None)
        additional_info = f" in type {default_type}" if default_type else ""
        exchange_connector.logger.warning(f"{exchange_connector.name} does not support sandboxing {additional_info}: {e}")
        # raise exception to stop this exchange and prevent dealing with a real funds exchange
        raise e
    return None


def load_markets_from_cache(client, market_filter: typing.Union[None, typing.Callable[[dict], bool]] = None):
    client.set_markets(
        market
        for market in ccxt_clients_cache.get_exchange_parsed_markets(ccxt_clients_cache.get_client_key(client))
        if market_filter is None or market_filter(market)
    )


def set_markets_cache(client):
    if client.markets:
        ccxt_clients_cache.set_exchange_parsed_markets(
            ccxt_clients_cache.get_client_key(client), copy.deepcopy(list(client.markets.values()))
        )


def get_ccxt_client_login_options(exchange_manager):
    """
    :return: ccxt client login option dict, can be overwritten to custom exchange login
    """
    if exchange_manager.is_future:
        return {'defaultType': 'future'}
    if exchange_manager.is_margin:
        return {'defaultType': 'margin'}
    return {'defaultType': 'spot'}


def get_symbols(client, active_only):
    try:
        if active_only:
            return set(
                symbol
                for symbol in client.symbols
                if client.markets.get(symbol, {}).get(
                    enums.ExchangeConstantsMarketStatusColumns.ACTIVE.value, True
                ) in (True, None)
            )
        return set(client.symbols)
    except (AttributeError, TypeError):
        # ccxt exchange load_markets failed
        return set()


def get_time_frames(client):
    try:
        if isinstance(client, ccxt_pro.Exchange):
            # ccxt pro exchanges might have different timeframes in options
            options_time_frames = client.safe_value(client.options, 'timeframes')
            if options_time_frames:
                values = set([
                    time_frame
                    for time_frame in options_time_frames
                    if time_frame_manager.is_time_frame(time_frame)
                ])
                if values:
                    return values
        # use normal client timeframes (values of rest exchange)
        return set(client.timeframes)
    except (AttributeError, TypeError):
        # ccxt exchange describe() is invalid
        return set()


def get_exchange_pair(client, pair) -> str:
    if pair in client.symbols:
        try:
            return client.market(pair)["id"]
        except KeyError:
            pass
    raise ValueError(f'{pair} is not supported')


def get_pair_cryptocurrency(client, pair) -> str:
    if pair in client.symbols:
        try:
            return client.market(pair)["base"]
        except KeyError:
            pass
    raise ValueError(f'{pair} is not supported')


def get_contract_size(client, pair) -> float:
    return client.markets[pair][ccxt_enums.ExchangeConstantsMarketStatusCCXTColumns.CONTRACT_SIZE.value]


def get_fees(market_status) -> dict:
    return {
        enums.ExchangeConstantsMarketPropertyColumns.TAKER.value:
            market_status.get(enums.ExchangeConstantsMarketPropertyColumns.TAKER.value,
                              constants.CONFIG_DEFAULT_FEES),
        enums.ExchangeConstantsMarketPropertyColumns.MAKER.value:
            market_status.get(enums.ExchangeConstantsMarketPropertyColumns.MAKER.value,
                              constants.CONFIG_DEFAULT_FEES),
        enums.ExchangeConstantsMarketPropertyColumns.FEE.value:
            market_status.get(enums.ExchangeConstantsMarketPropertyColumns.FEE.value,
                              constants.CONFIG_DEFAULT_FEES)
    }


def add_headers(client, headers_dict):
    """
    Add new headers to ccxt client
    :param headers_dict: the additional header keys and values as dict
    """
    for header_key, header_value in headers_dict.items():
        client.headers[header_key] = header_value


def add_options(client, options_dict):
    """
    Add new options to ccxt client
    :param options_dict: the additional option keys and values as dict
    """
    for option_key, option_value in options_dict.items():
        client.options[option_key] = option_value


def converted_ccxt_common_errors(f):
    async def converted_ccxt_common_errors_wrapper(*args, **kwargs):
        try:
            return await f(*args, **kwargs)
        except ccxt.RateLimitExceeded as err:
            raise errors.RateLimitExceeded(err) from err
        except ccxt.NotSupported as err:
            raise errors.NotSupported(err) from err
    return converted_ccxt_common_errors_wrapper


def _use_proxy_if_necessary(client, proxy_config: proxy_config_import.ProxyConfig):
    client.aiohttp_trust_env = proxy_config.aiohttp_trust_env
    if proxy_config.http_proxy:
        client.http_proxy = proxy_config.http_proxy
    if proxy_config.http_proxy_callback:
        client.http_proxy_callback = proxy_config.http_proxy_callback
    if proxy_config.https_proxy:
        client.https_proxy = proxy_config.https_proxy
    if proxy_config.https_proxy_callback:
        client.https_proxy_callback = proxy_config.https_proxy_callback
    if proxy_config.socks_proxy:
        client.socks_proxy = proxy_config.socks_proxy
    if proxy_config.socks_proxy_callback:
        client.socks_proxy_callback = proxy_config.socks_proxy_callback


def _get_client_config(
    options, headers, additional_config,
    api_key=None, secret=None, password=None, uid=None,
    auth_token=None, auth_token_header_prefix=None
):
    if auth_token is not None:
        headers["Authorization"] = f"{auth_token_header_prefix or ''}{auth_token}"
    config = {
        'verbose': constants.ENABLE_CCXT_VERBOSE,
        'enableRateLimit': constants.ENABLE_CCXT_RATE_LIMIT,
        'timeout': constants.DEFAULT_REQUEST_TIMEOUT,
        'options': options,
        'headers': headers
    }
    if api_key is not None:
        config['apiKey'] = api_key
    if secret is not None:
        config['secret'] = secret
    if password is not None:
        config['password'] = password
    if uid is not None:
        config['uid'] = uid
    config.update(additional_config or {})
    return config


def ccxt_exchange_class_factory(exchange_name):
    return getattr(async_ccxt, exchange_name)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

import pytest
from octobot_commons.constants import PORTFOLIO_AVAILABLE, PORTFOLIO_TOTAL

from octobot_trading.enums import TraderOrderType
from octobot_trading.personal_data.orders.types.limit.sell_limit_order import SellLimitOrder
from octobot_trading.personal_data.orders.types.market.buy_market_order import BuyMarketOrder
from octobot_trading.personal_data.portfolios.types.margin_portfolio import MarginPortfolio

from tests.exchanges import backtesting_trader, backtesting_config, backtesting_exchange_manager, fake_backtesting, \
    DEFAULT_EXCHANGE_NAME

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


@pytest.mark.parametrize("backtesting_exchange_manager", [(None, DEFAULT_EXCHANGE_NAME, False, True, False)],
                         indirect=["backtesting_exchange_manager"])
async def test_initial_portfolio(backtesting_trader):
    config, exchange_manager, trader = backtesting_trader
    portfolio_manager = exchange_manager.exchange_personal_data.portfolio_manager

    assert isinstance(portfolio_manager.portfolio, MarginPortfolio)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_commons.logging as logging


class OrderGroup:
    def __init__(self, name, orders_manager):
        self.name = name
        self.orders_manager = orders_manager
        self.logger = logging.get_logger(str(self))
        self.enabled = True

    async def on_fill(self, filled_order, ignored_orders=None):
        """
        Called when an order referencing this group is filled
        This is called right before updating portfolio for this filled order and the
        order fill publication
        :param filled_order: the filled order
        :param ignored_orders: orders that should be ignored
        """

    async def on_cancel(self, cancelled_order, ignored_orders=None):
        """
        Called when an order referencing this group is cancelled
        This is called before updating portfolio for this cancelled order and the
        order cancel publication
        :param cancelled_order: the cancelled order
        :param ignored_orders: orders that should be ignored
        """

    async def enable(self, enabled):
        self.enabled = enabled

    def get_group_open_orders(self):
        return [
            order
            for order in self.orders_manager.get_order_from_group(self.name)
            if order.is_open() and not order.is_cancelling()
        ]

    def clear(self):
        self.orders_manager = None

    def __str__(self):
        return f"{self.__class__.__name__} #{self.name}"

    def __eq__(self, other):
        return self.__class__ is other.__class__ and \
               self.orders_manager is other.orders_manager \
               and self.name == other.name



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_commons.channels_name as channels_names
import async_channel.util as channel_creator
import async_channel.channels as channels
import octobot_trading.signals.channel.remote_trading_signal as remote_trading_signal
import octobot_trading.signals.channel.signal_producer as signal_producer


async def create_remote_trading_signal_channel_if_missing(exchange_manager) -> \
        (remote_trading_signal.RemoteTradingSignalsChannel, bool):
    try:
        return channels.get_chan(channels_names.OctoBotCommunityChannelsName.REMOTE_TRADING_SIGNALS_CHANNEL.value), \
               False
    except KeyError:
        channel = await channel_creator.create_channel_instance(remote_trading_signal.RemoteTradingSignalsChannel,
                                                                channels.set_chan)
        # also create the associated producer
        producer = signal_producer.RemoteTradingSignalProducer(
            channel,
            exchange_manager.bot_id
        )
        await channel.register_producer(producer)
        return channel, True



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal

import octobot_trading.constants as constants
import octobot_trading.enums as enums
import octobot_trading.personal_data.positions.position as position_class


class LinearPosition(position_class.Position):
    def update_value(self):
        """
        Notional value = CONTRACT_QUANTITY * MARK_PRICE
        """
        self.value = self.size * self.mark_price

    def get_unrealized_pnl(self, price):
        """
        LONG_PNL = CONTRACT_QUANTITY x [MARK_PRICE - ENTRY_PRICE]
        SHORT_PNL = CONTRACT_QUANTITY x [ENTRY_PRICE - MARK_PRICE]
        :param price: the pnl calculation price
        :return: the unrealized pnl
        """
        # ensure update validity
        if price <= constants.ZERO or self.entry_price <= constants.ZERO:
            return constants.ZERO
        if self.is_long():
            return self.size * (price - self.entry_price)
        if self.is_short():
            return -self.size * (self.entry_price - price)
        return constants.ZERO

    def get_margin_from_size(self, size):
        """
        Calculates margin from size : margin = (Position quantity x entry price) / leverage
        """
        return (size * self.entry_price) / self.symbol_contract.current_leverage

    def get_size_from_margin(self, margin):
        """
        Calculates size from margin : size = (margin x leverage) / entry price
        """
        return (margin * self.symbol_contract.current_leverage) / self.entry_price

    def calculate_maintenance_margin(self):
        """
        :return: Maintenance margin = Position quantity x entry price x Maintenance margin rate
        """
        return self.size * self.entry_price * self.symbol_contract.maintenance_margin_rate

    def update_isolated_liquidation_price(self):
        """
        Updates isolated position liquidation price
        LONG LIQUIDATION PRICE = ENTRY_PRICE * (1 - Initial Margin Rate + MAINTENANCE_MARGIN_RATE)
        SHORT LIQUIDATION PRICE = ENTRY_PRICE * (1 + Initial Margin Rate - MAINTENANCE_MARGIN_RATE)
        - Long : - Extra Margin Added/ Contract Size
        - Short : + Extra Margin Added/ Contract Size
        """
        try:
            if self.is_long():
                self.liquidation_price = self.entry_price * (
                        constants.ONE - self.get_initial_margin_rate() + self.symbol_contract.maintenance_margin_rate)
            elif self.is_short():
                self.liquidation_price = self.entry_price * (
                        constants.ONE + self.get_initial_margin_rate() - self.symbol_contract.maintenance_margin_rate)
            else:
                self.liquidation_price = constants.ZERO
            self.update_fee_to_close()
        except (decimal.DivisionByZero, decimal.InvalidOperation):
            self.liquidation_price = constants.ZERO

    def get_bankruptcy_price(self, price, side, with_mark_price=False):
        """
        :param price: the price to compute bankruptcy from
        :param side: the side of the position
        :param with_mark_price: if price should be mark price instead of entry price
        :return: Bankruptcy Price
        Long position = Entry Price x (1 - Initial Margin Rate)
        Short position = Entry Price × (1 + Initial Margin Rate)
        """
        if side is enums.PositionSide.LONG:
            return self.mark_price if with_mark_price else \
                price * (constants.ONE - self.get_initial_margin_rate())
        elif side is enums.PositionSide.SHORT:
            return self.mark_price if with_mark_price else \
                price * (constants.ONE + self.get_initial_margin_rate())
        return constants.ZERO

    def get_fee_to_open(self, quantity, price, symbol):
        """
        :return: Fee to open = (Quantity * Mark Price) x Taker fee
        """
        return quantity * price * self.get_taker_fee(symbol)

    def get_fee_to_close(self, quantity, price, side, symbol, with_mark_price=False):
        """
        :return: Fee to open = (Quantity * Mark Price) x Taker fee
        """
        return quantity * self.get_bankruptcy_price(price, side, with_mark_price=with_mark_price) * \
            self.get_taker_fee(symbol)

    def get_order_cost(self):
        """
        :return: Order Cost = Initial Margin + Two-Way Taker Fee
        """
        return self.initial_margin + self.get_two_way_taker_fee()

    def update_fee_to_close(self):
        """
        :return: Fee to close = (Quantity * Bankruptcy Price derived from mark price) x Taker fee
        """
        self.fee_to_close = self.get_fee_to_close(self.size, self.entry_price, self.side, self.symbol,
                                                  with_mark_price=True)

    def update_average_entry_price(self, update_size, update_price):
        """
        Average entry price = total contract value in market / total quantity of contracts
        Total contract value in market = [(Current position quantity * Current position entry price)
                                          + (Update quantity * Update price)]
        """
        total_contract_value = self.size + update_size
        self.entry_price = ((self.size * self.entry_price + update_size * update_price) /
                            (total_contract_value if total_contract_value != constants.ZERO else constants.ONE))
        if self.entry_price < constants.ZERO:
            self.entry_price = constants.ZERO

    def update_average_exit_price(self, update_size, update_price):
        """
        Average exit price = total contract value in market / total quantity of contracts
        Total contract value in market = [(Current position quantity * Current position exit price)
                                          + (Update quantity * Update price)]
        """
        if self.exit_price == constants.ZERO:
            self.exit_price = update_price
        else:
            total_contract_value = self.already_reduced_size + update_size
            self.exit_price = ((self.already_reduced_size * self.exit_price + update_size * update_price) /
                               (total_contract_value if total_contract_value != constants.ZERO else constants.ONE))
        if self.exit_price < constants.ZERO:
            self.exit_price = constants.ZERO

    @staticmethod
    def is_inverse():
        return False



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import abc
import contextlib
import decimal

import octobot_commons.constants as common_constants
import octobot_commons.enums as common_enums
import octobot_commons.logging as logging
import octobot_commons.tentacles_management as abstract_tentacle

import async_channel.constants as channel_constants
import async_channel.channels as channels

import octobot_tentacles_manager.api as tentacles_manager_api
import octobot_tentacles_manager.configuration as tm_configuration

import octobot_trading.constants as constants
import octobot_trading.enums as enums
import octobot_trading.exchange_channel as exchanges_channel
import octobot_trading.modes.modes_factory as modes_factory
import octobot_trading.modes.channel.abstract_mode_producer as abstract_mode_producer
import octobot_trading.modes.channel.abstract_mode_consumer as abstract_mode_consumer
import octobot_trading.modes.mode_config as mode_config
import octobot_trading.modes.modes_util as modes_util
import octobot_trading.exchanges.util.exchange_util as exchange_util
import octobot_trading.signals as signals


class AbstractTradingMode(abstract_tentacle.AbstractTentacle):
    __metaclass__ = abc.ABCMeta
    USER_INPUT_TENTACLE_TYPE = common_enums.UserInputTentacleTypes.TRADING_MODE
    ALLOW_CUSTOM_TRIGGER_SOURCE = False
    HISTORIZE_USER_INPUT_CONFIG = True

    MODE_PRODUCER_CLASSES = []
    MODE_CONSUMER_CLASSES = []
    # maximum seconds before sending a trading signal if orders are slow to create on exchange
    TRADING_SIGNAL_TIMEOUT = 10
    REQUIRE_TRADES_HISTORY = False   # set True when this trading mode needs the trade history to operate
    ALLOW_CANCEL_BEFORE_BUY_SIGNALS = True  # set False if trade signals from this trading mode should NOT be
    # reordered to process cancel signals first
    SUPPORTS_INITIAL_PORTFOLIO_OPTIMIZATION = False  # set True when self._optimize_initial_portfolio is implemented
    SUPPORTS_HEALTH_CHECK = False   # set True when self.health_check is implemented
    ENABLE_HEALTH_CHECK = "enable_health_check"
    HEALTH_CHECK_INTERVAL = common_constants.DAYS_TO_SECONDS

    def __init__(self, config, exchange_manager):
        super().__init__()
        self.logger = logging.get_logger(self.get_name())

        # Global OctoBot configuration
        self.config: dict = config

        # Mode related exchange manager instance
        self.exchange_manager = exchange_manager

        # The id of the OctoBot using this trading mode
        self.bot_id: str = None

        # Trading Mode specific config (Is loaded from tentacle specific file)
        self.trading_config: dict = None

        # Trading Mode specific config snapshot before a config update
        self.previous_trading_config: dict = None

        # If this mode is enabled
        self.enabled: bool = True

        # Specified Cryptocurrency for this instance (Should be None if wildcard)
        self.cryptocurrency: str = None

        # Symbol is the cryptocurrency pair (Should be None if wildcard)
        self.symbol: str = None

        # Time_frame is the chart time frame (Should be None if wildcard)
        self.time_frame = None

        # producers is the list of producers created by this trading mode
        self.producers = []

        # producers is the list of consumers created by this trading mode
        self.consumers = []

        # True when this trading mode is waken up only after full candles close
        self.is_triggered_after_candle_close = False

        # True when initialization orders are waiting to be created
        self.are_initialization_orders_pending = False

        # When True, health check will be performed when calling trading_mode_trigger
        self.is_health_check_enabled = False
        self._last_health_check_time = 0

    # Used to know the current state of the trading mode.
    # Overwrite in subclasses
    def get_current_state(self) -> tuple:
        """
        :return: (str, float): (current state description, current state value)
        """
        return "N/A", 0

    @classmethod
    def get_is_cryptocurrency_wildcard(cls) -> bool:
        """
        :return: True if the mode is not cryptocurrency dependant else False
        """
        return True

    @classmethod
    def get_is_symbol_wildcard(cls) -> bool:
        """
        :return: True if the mode is not symbol dependant else False
        """
        return True

    @classmethod
    def get_is_time_frame_wildcard(cls) -> bool:
        """
        :return: True if the mode is not time_frame dependant else False
        """
        return True

    @classmethod
    def get_supported_exchange_types(cls) -> list:
        """
        :return: The list of supported exchange types
        """
        return [
            enums.ExchangeTypes.SPOT
        ]

    def get_mode_producer_classes(self) -> list:
        return self.MODE_PRODUCER_CLASSES

    def get_mode_consumer_classes(self) -> list:
        return self.MODE_CONSUMER_CLASSES

    def should_emit_trading_signal(self) -> bool:
        """
        :return: True if the mode should be emitting trading signals according to configuration and trading environment
        """
        return not self.exchange_manager.is_backtesting and mode_config.is_trading_signal_emitter(self)

    def get_trading_signal_identifier(self) -> str:
        """
        :return: The identifier of the trading signal from config or the name of the tentacle if missing
        """
        try:
            return self.trading_config[common_constants.CONFIG_TRADING_SIGNALS_STRATEGY] or self.get_name()
        except KeyError:
            return self.get_name()

    def is_following_trading_signals(self) -> bool:
        """
        :return: True when the trading mode is following trading signals
        """
        return False

    @classmethod
    def get_is_trading_on_exchange(cls, exchange_name,
                                   tentacles_setup_config: tm_configuration.TentaclesSetupConfiguration) -> bool:
        """
        :return: When returning false, the associated exchange_manager.is_trading will be set to false, which will
        prevent the initialization of trade related elements. Default is True
        """
        return True

    @classmethod
    def is_ignoring_cancelled_orders_trades(cls) -> bool:
        """
        :return: True if trades created by cancelled orders should be ignored and not stored in trades manager
        """
        return False

    @classmethod
    def get_parent_trading_mode_classes(cls, higher_parent_class_limit=None) -> list:
        return [
            class_type
            for class_type in cls.mro()
            if (higher_parent_class_limit if higher_parent_class_limit else AbstractTradingMode) in class_type.mro()
        ]

    @staticmethod
    def is_backtestable() -> bool:
        """
        Should be overwritten
        :return: True if the TradingMode can be used in a backtesting else False
        """
        return True

    async def initialize(self, trading_config=None, auto_start=True) -> None:
        """
        Triggers producers and consumers creation
        """
        await self.reload_config(self.exchange_manager.bot_id, trading_config=trading_config)
        self.producers = await self.create_producers(auto_start)
        self.consumers = await self.create_consumers()

    async def stop(self) -> None:
        """
        Stops all producers and consumers
        """
        for producer in self.producers:
            await producer.stop()
        for consumer in self.consumers:
            await consumer.stop()
        self.exchange_manager = None

    async def create_producers(self, auto_start) -> list:
        """
        Creates the instance of producers listed in MODE_PRODUCER_CLASSES
        :return: the list of producers created
        """
        return [
            await self._create_mode_producer(mode_producer_class, auto_start)
            for mode_producer_class in self.get_mode_producer_classes()
        ]

    async def _create_mode_producer(self, mode_producer_class, auto_start):
        """
        Creates a new :mode_producer_class: instance and starts it
        :param mode_producer_class: the trading mode producer class to create
        :return: the producer class created
        """
        mode_producer = mode_producer_class(
            exchanges_channel.get_chan(constants.MODE_CHANNEL, self.exchange_manager.id),
            self.config, self, self.exchange_manager)
        if auto_start:
            await mode_producer.run()
        return mode_producer

    async def start_producers(self):
        """
        Should be used if producers got created with auto_start=False
        """
        for producer in self.producers:
            await producer.run()

    async def create_consumers(self) -> list:
        """
        Creates the instance of consumers listed in MODE_CONSUMER_CLASSES
        :return: the list of consumers created
        """
        base_consumers = [
            await self._create_mode_consumer(mode_consumer_class)
            for mode_consumer_class in self.get_mode_consumer_classes()
        ]
        if user_input_consumer := await self._create_user_input_consumer():
            base_consumers.append(user_input_consumer)

        return base_consumers

    async def _create_user_input_consumer(self):
        try:
            import octobot_services.channel as services_channels
            user_commands_consumer = \
                await channels.get_chan(services_channels.UserCommandsChannel.get_name()).new_consumer(
                    self.user_commands_callback,
                    {"bot_id": self.bot_id, "subject": self.get_name()}
                )
            return user_commands_consumer
        except KeyError:
            self.logger.debug(f"{services_channels.UserCommandsChannel.get_name()} unavailable")
        except ImportError:
            self.logger.warning("Can't connect to services channels")
        return None

    async def user_commands_callback(self, bot_id, subject, action, data) -> None:
        self.logger.debug(f"Received {action} command")
        if action == common_enums.UserCommands.MANUAL_TRIGGER.value:
            self.logger.debug(f"Triggering trading mode from {action} command with data: {data}")
            await self._manual_trigger(data)
        if action == common_enums.UserCommands.RELOAD_CONFIG.value:
            await self.reload_config(bot_id)
            self.logger.debug("Reloaded configuration")
        elif action == common_enums.UserCommands.CLEAR_PLOTTING_CACHE.value:
            await modes_util.clear_plotting_cache(self)
        elif action == common_enums.UserCommands.CLEAR_SIMULATED_ORDERS_CACHE.value:
            await modes_util.clear_simulated_orders_cache(self)
        elif action == common_enums.UserCommands.OPTIMIZE_INITIAL_PORTFOLIO.value:
            if self.SUPPORTS_INITIAL_PORTFOLIO_OPTIMIZATION:
                await self.optimize_initial_portfolio([], {})
        elif action == common_enums.UserCommands.TRIGGER_HEALTH_CHECK.value:
            if self.SUPPORTS_HEALTH_CHECK:
                await self.health_check([], {})

    async def _manual_trigger(self, data):
        kwargs = {
            "trigger_source": common_enums.TriggerSource.MANUAL.value
        }
        kwargs.update(data.get("kwargs", {}))
        for producer in self.producers:
            await producer.trigger(**kwargs)

    def enabled_health_check_in_config(self) -> bool:
        try:
            return self.trading_config.get(self.ENABLE_HEALTH_CHECK, False)
        except AttributeError:
            # when self.trading_config is None, should not happen
            return False

    @classmethod
    def get_forced_updater_channels(cls) -> set[str]:
        return set()

    def _health_check_interval_expired(self) -> bool:
        return self.exchange_manager.exchange.get_exchange_current_time() - self._last_health_check_time \
            > self.HEALTH_CHECK_INTERVAL

    def is_health_check_required(self) -> bool:
        return self.SUPPORTS_HEALTH_CHECK and self.is_health_check_enabled and self._health_check_interval_expired()

    async def health_check(self, chained_orders: list, tickers: dict) -> list:
        self._last_health_check_time = self.exchange_manager.exchange.get_exchange_current_time()
        if not self.producers:
            # nothing to do
            return []
        async with self._single_exchange_operation("health check") as continue_operation:
            if not continue_operation:
                return []
            return await self.single_exchange_process_health_check(chained_orders, tickers)

    async def single_exchange_process_health_check(self, chained_orders: list, tickers: dict) -> list:
        raise NotImplementedError("single_exchange_process_health_check is not implemented")

    async def optimize_initial_portfolio(self, sellable_assets: list, tickers: dict) -> list:
        if not self.producers:
            # nothing to do
            return []
        # first acquire trading mode lock to be sure we are not in during trading mode iteration
        async with self.producers[0].trading_mode_trigger():
            async with self._single_exchange_operation("portfolio optimization") as continue_operation:
                if not continue_operation:
                    return []
                target_asset = exchange_util.get_common_traded_quote(self.exchange_manager)
                if target_asset is None:
                    self.logger.error(f"Impossible to optimize initial portfolio with different quotes in traded pairs")
                    return []
                self.logger.info(f"Starting portfolio optimization using trading mode with symbol {self.symbol}")
                created_orders = await self.single_exchange_process_optimize_initial_portfolio(
                    sellable_assets, target_asset, tickers
                )
                if not created_orders:
                    self.logger.info("Optimizing portfolio: no order to create")
                await modes_util.notify_portfolio_optimization_complete()
                return created_orders

    async def single_exchange_process_optimize_initial_portfolio(
        self, sellable_assets: list, target_asset: str, tickers: dict
    ) -> list:
        raise NotImplementedError("single_exchange_process_optimize_initial_portfolio is not implemented")

    @contextlib.asynccontextmanager
    async def _single_exchange_operation(self, operation_name):
        is_the_single_operation = not self.producers[0].producer_exchange_wide_lock(self.exchange_manager).locked()
        if is_the_single_operation:
            # already locked by another trading mode instance: this other trading mode will do the operation
            self.logger.debug(
                f"[Single exchange operation]: triggering '{operation_name}' for trading mode with symbol "
                f"{self.symbol} [{self.exchange_manager.exchange_name}]."
            )
        else:
            # already locked by another trading mode instance: this other trading mode will do the operation
            self.logger.debug(
                f"[Single exchange operation]: skipping '{operation_name}' for trading mode with symbol {self.symbol} "
                f"[{self.exchange_manager.exchange_name}]: "
                f"{operation_name} already in progress. Waiting for initial {operation_name} to complete."
            )
        async with self.producers[0].producer_exchange_wide_lock(self.exchange_manager):
            yield is_the_single_operation

    @classmethod
    def get_user_commands(cls) -> dict:
        """
        Return the dict of user commands for this tentacle
        :return: the commands dict
        """
        commands = {}
        if cls.SUPPORTS_INITIAL_PORTFOLIO_OPTIMIZATION:
            commands[common_enums.UserCommands.OPTIMIZE_INITIAL_PORTFOLIO.value] = {}
        if cls.SUPPORTS_HEALTH_CHECK:
            commands[common_enums.UserCommands.TRIGGER_HEALTH_CHECK.value] = {}
        return commands

    async def _create_mode_consumer(self, mode_consumer_class):
        """
        Creates a new :mode_consumer_class: instance and subscribe this new consumer to the trading mode channel
        :param mode_consumer_class: the trading mode consumer class to create
        :return: the consumer class created
        """
        mode_consumer = mode_consumer_class(self)
        await exchanges_channel.get_chan(constants.MODE_CHANNEL, self.exchange_manager.id).new_consumer(
            consumer_instance=mode_consumer,
            trading_mode_name=self.get_name(),
            cryptocurrency=self.cryptocurrency if self.cryptocurrency else channel_constants.CHANNEL_WILDCARD,
            symbol=self.symbol if self.symbol else channel_constants.CHANNEL_WILDCARD,
            time_frame=self.time_frame if self.time_frame else channel_constants.CHANNEL_WILDCARD)
        return mode_consumer

    async def reload_config(self, bot_id: str, trading_config=None) -> None:
        """
        Try to load TradingMode tentacle config.
        Calls set_default_config() if the tentacle config is empty
        """
        self.trading_config = trading_config or \
            tentacles_manager_api.get_tentacle_config(self.exchange_manager.tentacles_setup_config, self.__class__)
        # set default config if nothing found
        if not self.trading_config:
            self.set_default_config()
        await self.load_and_save_user_inputs(bot_id)
        for element in self.consumers + self.producers:
            if isinstance(element, (abstract_mode_consumer.AbstractTradingModeConsumer,
                                    abstract_mode_producer.AbstractTradingModeProducer)):
                element.on_reload_config()
                await element.init_user_inputs(False)
        self.logger.debug(f"Using config: {self.trading_config}")

    def get_local_config(self):
        return self.trading_config

    @classmethod
    def create_local_instance(cls, config, tentacles_setup_config, tentacle_config):
        return modes_factory.create_temporary_trading_mode_with_local_config(
            cls, config, tentacle_config
        )

    # to implement in subclasses if config is necessary
    def set_default_config(self) -> None:
        pass

    """
    Strategy related methods
    """

    @classmethod
    def get_required_strategies_names_and_count(
        cls, tentacles_config: tm_configuration.TentaclesSetupConfiguration, trading_mode_config=None
    ):
        config = trading_mode_config or tentacles_manager_api.get_tentacle_config(tentacles_config, cls)
        if constants.TRADING_MODE_REQUIRED_STRATEGIES in config:
            return config[constants.TRADING_MODE_REQUIRED_STRATEGIES], cls.get_required_strategies_count(config)
        raise Exception(f"'{constants.TRADING_MODE_REQUIRED_STRATEGIES}' is missing in configuration file")

    @classmethod
    def get_default_strategies(cls,
                               tentacles_config: tm_configuration.TentaclesSetupConfiguration,
                               trading_mode_config=None):
        config = trading_mode_config or tentacles_manager_api.get_tentacle_config(tentacles_config, cls)
        if common_constants.TENTACLE_DEFAULT_CONFIG in config:
            return config[common_constants.TENTACLE_DEFAULT_CONFIG]

        strategies_classes, _ = cls.get_required_strategies_names_and_count(tentacles_config, config)
        return strategies_classes

    @classmethod
    def get_required_strategies_count(cls, config):
        min_strategies_count = 1
        if constants.TRADING_MODE_REQUIRED_STRATEGIES_MIN_COUNT in config:
            min_strategies_count = config[constants.TRADING_MODE_REQUIRED_STRATEGIES_MIN_COUNT]
        return min_strategies_count

    @classmethod
    def get_required_candles_count(cls, tentacles_setup_config: tm_configuration.TentaclesSetupConfiguration):
        return tentacles_manager_api.get_tentacle_config(tentacles_setup_config, cls).get(
            constants.CONFIG_CANDLES_HISTORY_SIZE_KEY,
            common_constants.DEFAULT_IGNORED_VALUE
        )

    @contextlib.asynccontextmanager
    async def remote_signal_publisher(self, symbol: str):
        async with signals.remote_signal_publisher(self.exchange_manager, symbol, self.should_emit_trading_signal()) \
             as signal_builder:
            yield signal_builder

    async def create_order(self, order, loaded: bool = False, params: dict = None,
                           wait_for_creation=True,
                           creation_timeout=constants.INDIVIDUAL_ORDER_SYNC_TIMEOUT):
        return await signals.create_order(
            self.exchange_manager, self.should_emit_trading_signal(), order,
            loaded=loaded, params=params,
            wait_for_creation=wait_for_creation, creation_timeout=creation_timeout
        )

    async def cancel_order(self, order, ignored_order: object = None) -> bool:
        return await signals.cancel_order(
            self.exchange_manager, self.should_emit_trading_signal(), order,
            ignored_order=ignored_order
        )

    async def cancel_all_orders(
        self,
        symbol: str,
        allow_single_order_cancel_fallback: bool,
        wait_for_cancelling: bool = True,
        cancelling_timeout: float = constants.INDIVIDUAL_ORDER_SYNC_TIMEOUT
    ) -> bool:
        # todo send signals when implementation will be necessary
        return await self.exchange_manager.trader.cancel_all_orders(
            symbol,
            allow_single_order_cancel_fallback,
            wait_for_cancelling=wait_for_cancelling,
            cancelling_timeout=cancelling_timeout
        )

    async def edit_order(self, order,
                         edited_quantity: decimal.Decimal = None,
                         edited_price: decimal.Decimal = None,
                         edited_stop_price: decimal.Decimal = None,
                         edited_current_price: decimal.Decimal = None,
                         params: dict = None) -> bool:
        return await signals.edit_order(
            self.exchange_manager, self.should_emit_trading_signal(), order,
            edited_quantity=edited_quantity,
            edited_price=edited_price,
            edited_stop_price=edited_stop_price,
            edited_current_price=edited_current_price,
            params=params
        )

    async def get_additional_metadata(self, is_backtesting):
        """
        Override if necessary
        """
        return {}

    def flush_trading_mode_consumers(self):
        for consumer in self.get_trading_mode_consumers():
            consumer.flush()

    def get_trading_mode_consumers(self):
        return [
            consumer
            for consumer in self.consumers
            if isinstance(consumer, abstract_mode_consumer.AbstractTradingModeConsumer)
        ]
